from __future__ import annotations

import logging
import time
from pathlib import Path
from typing import Any
import shutil

from phenomate_core.preprocessing.base import BasePreprocessor

# shared_logger = logging.getLogger("celery")
from phenomate_core.get_version import get_task_logger
shared_logger = get_task_logger(__name__)

class ImuPreprocessor(BasePreprocessor[Path]):
    """
    Code based upon Resonate IMU writer: resonatesystems-rs24005-appn-instrument-interfaces-29_09_2005/Instruments/IMU
    - The Pheomate system currently collects the raw IMU data as a .bin file (which notably is not a protobuffer 
    defined file) and then extracts the two types of entry, depending on the decoded message id and adds the 
    amiga_timestamp of when the data was added to the CSV. Message id 2561 is saved to the same filename as the `.bin`
    data, but with a `.csv` extension, and message id 2562 is saved with a `_GNSS.csv` extension.

    The IMU being used is an ACEINNA INS401 GNSS/RTK see: https://www.aceinna.com/inertial-systems/INS401
    User manual: 7430-4006-02_B_D8_INS401_UserManual.pdf
    
    There are three message ids and the data they contain are listed in the next 3 tables.

    1. High time resolution raw data: amiga_timestamp,fgps_week,gps_millis,ax,ay,az,gx,gy,gz

    There is a JSON-LD CSVW description for this data in file:  imu_raw_observations.csv-metadata.json and imu_observations.csv-metadata.json

    **Table 11 from INS401 user guide: Raw IMU Data**
    Message ID     : 0x0a01 (2561)
    Message Length : 30 bytes
    Message rate   : 1000 Hz
    | **Field**       | **Offset (Byte)** | **Type** | **Unit**   | **Description**             |
    |-----------------|-------------------|----------|------------|-----------------------------|
    | `gps_week`      | 0                 | `uint16` |            | GPS week                    |
    | `gps_millisecs` | 2                 | `uint32` | ms         | GPS time of week            |
    | `accel_x`       | 6                 | `float`  | m/s²       | Acceleration on X axis      |
    | `accel_y`       | 10                | `float`  | m/s²       | Acceleration on Y axis      |
    | `accel_z`       | 14                | `float`  | m/s²       | Acceleration on Z axis      |
    | `gyro_x`        | 18                | `float`  | deg/s      | Angular rate on X axis      |
    | `gyro_y`        | 22                | `float`  | deg/s      | Angular rate on Y axis      |
    | `gyro_z`        | 26                | `float`  | deg/s      | Angular rate on Z axis      |


    2. Lower time resolution GNS processed data: amiga_timestamp,gps_week,gps_millisecs,position_type,
                                                 latitude,longitude,height,latitude_std,
                                                 longitude_std,height_std,numberOfSVs,numberOfSVs_in_solution,
                                                 hdop,diffage,north_vel,east_vel,up_vel,north_vel_std,
                                                 east_vel_std,up_vel_std

    There is a JSON-LD CSVW description for this data in file:  gnss_observations.csv-metadata.json

    **Table 4 from INS401 user guide: GNSS Solution Packet** from the INS401 User Manual
    Message ID     : 0x0a02 (2562)
    Message Length : 77 bytes
    Message rate   : 1 Hz
    | **Field**               | **Offset (Byte)** | **Type**   | **Unit** | **Description**                                                                 |
    |-------------------------|-------------------|------------|----------|---------------------------------------------------------------------------------|
    | `gps_week`              | 0                 | `uint16`   |          | GPS week                                                                       |
    | `gps_millisecs`         | 2                 | `uint32`   | ms       | GPS time of week                                                               |
    | `position_type`         | 6                 | `uint8`    |          | Positioning type: 0: INVALID 1: SPP ==2: RTD 3: INS_PROPAGATED 4: RTK_FIXED 5: RTK_FLOAT |
    | `latitude`              | 7                 | `double`   | deg      | Geodetic latitude                                                              |
    | `longitude`             | 15                | `double`   | deg      | Geodetic longitude                                                             |
    | `height`                | 23                | `double`   | m        | Height above ellipsoid                                                         |
    | `latitude_std`          | 31                | `float`    | m        | Latitudinal position accuracy                                                  |
    | `longitude_std`         | 35                | `float`    | m        | Longitudinal position accuracy                                                 |
    | `height_std`            | 39                | `float`    | m        | Vertical position accuracy                                                     |
    | `numberOfSVs`           | 43                | `uint8`    |          | Number of satellites                                                           |
    | `numberOfSVs_in_solution` | 44              | `uint8`    |          | Number of satellites used in solution                                          |
    | `Hdop`                  | 45                | `float`    |          | Horizontal Dilution of Precision                                               |
    | `Diffage`               | 49                | `float`    | s        | Age of differential GNSS correction                                            |
    | `north_vel`             | 53                | `float`    | m/s      | North velocity                                                                 |
    | `east_vel`              | 57                | `float`    | m/s      | East velocity                                                                  |
    | `up_vel`                | 61                | `float`    | m/s      | Up velocity                                                                    |
    | `north_vel_std`         | 65                | `float`    | m/s      | North velocity accuracy                                                        |
    | `east_vel_std`          | 69                | `float`    | m/s      | East velocity accuracy                                                         |
    | `up_vel_std`            | 73                | `float`    | m/s      | Up velocity accuracy                                                           |



    3. This data is not currently output
    Table 5 from the INS401 User Manual, formatted in Markdown.

    There is a JSON-LD CSVW description for this data in file:  ins_100hz_observations.csv-metadata.json

    **Table 5 from INS401 user guide7: INS Solution Packet**
    Message ID     : 0x0a03 (2563)
    Message Length :110 bytes
    Message rate   : 100 Hz

    Requirements for the data output:
    - There is no significant steering action of the vehicle (Z-gyro rate <+/-5 dps)
    - Speed > 10km/h
    - 3s is needed if RTK solution is available
    | **Field**               | **Offset (Byte)** | **Type**   | **Unit** | **Description**                                                                 |
    |------------------------|-------------------|------------|----------|---------------------------------------------------------------------------------|
    | `gps_week`             | 0                 | `uint16`   |          | GPS week                                                                       |
    | `gps_millisecs`        | 2                 | `uint32`   | ms       | GPS time of week                                                               |
    | `ins_status`           | 6                 | `uint8`    |          | INS status: 0: INVALID 1: INS_ALIGNING 2: INS_HIGH_VARIANCE 3: INS_SOLUTION_GOOD 4: INS_SOLUTION_FREE 5: INS_ALIGNMENT_COMPLETE |
    | `ins_position_type`    | 7                 | `uint8`    |          | INS position type: 0: INVALID 1: SPP/INS 2: RTD/INS 3: INS_PROPAGATE 4: RTK_FIXED/INS 5: RTK_FLOAT/INS |
    | `latitude`             | 8                 | `double`   | deg      | Geodetic latitude                                                              |
    | `longitude`            | 16                | `double`   | deg      | Geodetic longitude                                                             |
    | `height`               | 24                | `double`   | m        | Height above ellipsoid                                                         |
    | `north_velocity`       | 32                | `float`    | m/s      | North velocity in navigation ENU frame                                         |
    | `east_velocity`        | 36                | `float`    | m/s      | East velocity in navigation ENU frame                                          |
    | `up_velocity`          | 40                | `float`    | m/s      | Up velocity in navigation ENU frame                                            |
    | `longitudinal_velocity`| 44                | `float`    | m/s      | Forward velocity in vehicle frame                                              |
    | `lateral_velocity`     | 48                | `float`    | m/s      | Lateral velocity in vehicle frame                                              |
    | `roll`                 | 52                | `float`    | deg      | Vehicle roll                                                                   |
    | `pitch`                | 56                | `float`    | deg      | Vehicle pitch                                                                  |
    | `heading`              | 60                | `float`    | deg      | Vehicle heading                                                                |
    | `latitude_std`         | 64                | `float`    | m        | Latitudinal position accuracy                                                  |
    | `longitude_std`        | 68                | `float`    | m        | Longitudinal position accuracy                                                 |
    | `height_std`           | 72                | `float`    | m        | Vertical position accuracy                                                     |
    | `north_velocity_std`   | 76                | `float`    | m/s      | North velocity accuracy                                                        |
    | `east_velocity_std`    | 80                | `float`    | m/s      | East velocity accuracy                                                         |
    | `up_velocity_std`      | 84                | `float`    | m/s      | Up velocity accuracy                                                           |
    | `long_vel_std`         | 88                | `float`    | m/s      | Longitudinal velocity accuracy                                                 |
    | `lat_vel_std`          | 92                | `float`    | m/s      | Lateral velocity accuracy                                                      |
    | `roll_std`             | 96                | `float`    | deg      | Vehicle roll accuracy                                                          |
    | `pitch_std`            | 100               | `float`    | deg      | Vehicle pitch accuracy                                                         |
    | `heading_std`          | 104               | `float`    | deg      | Vehicle heading accuracy                                                       |
    | `continent_id`         | 108               | `int16`    |          | Continent ID: -2: NONE -1: ERROR 0: UNKNOWN 1: ASIA 2: EUROPE 3: OCEANIA 4: AFRICA 5: NORTH AMERICA 6: SOUTH AMERICA 7: ANTARCTICA |


    """

    def extract(self, **kwargs: Any) -> None:
        """
        This method reads the '.origin' file extended version of the .bin file selected and
        reads the source directory (the directory where the .bin file was selected in the
        Phenomate GUI) and looks for files of the same name and timestamp to select and stores
        this as a list of filenames in the self.images list, to be processed in the save() method.
        """

        dir_part = self.path.parent  # this is another Path type
        file_part = self.path.name  # this is a str

        # # Read the path that was written to the self.path+'.origin' file        
        # origin_path = self.open_origin_file()  
        # # Select the matching files from the path by the timestamp of the self.path file
        # path_objects = self.matched_file_list(origin_path, file_part)
        
        # self.extra_files = path_objects
        # shared_logger.info(f"IMU data transfer: number of related files:  {len(self.extra_files)}")
        # self.images = path_objects

    def copy_extra_files(self, fpath: Path) -> None:
        """
        Extra files that are associated with the .bin proto buffer data can be copied to the destination directory.
        
        :fpath Path: is the directory in which to save the files
        
        This method impicitly uses the extra_files list that should be populated in the extract() method using
        open_origin_file() and matched_file_list() (see: ImuPreprocessor.extract())
        
        There are currently two specialised file copying sections to this 
        
        """
        # for file_path in self.extra_files:
            # try:
                # # For the IMU there should be 3 files, 1 bin file and two CSV files
                # # and one CSV has GNSS in the filename
                # if file_path.suffix == ".bin":
                    # file_path_name_ext = fpath / self.get_output_name(
                        # index=None, ext="bin", details=None
                    # )
                # elif file_path.suffix.lower() == ".csv":
                    # file_path = Path(file_path)
                    # output_file = file_path.name  # str
                    # # As the output filename isderived from the input .bin file, we need to add back the
                    # # the GNSS part to the GNSS CSV filename
                    # if output_file.lower().find("_gnss.csv") >= 0:
                        # file_path_name_ext = fpath / self.get_output_name(
                            # index=None, ext="csv", details="GNSS"
                        # )
                    # else:
                        # file_path_name_ext = fpath / self.get_output_name(
                            # index=None, ext="csv", details=None
                        # )
                # shutil.copy(file_path, file_path_name_ext)
                # shared_logger.info(f"BasePreprocessor.copy_extra_files(): IMU data transfer: Copied file: {file_path_name_ext}")

            # except FileNotFoundError as e:
                # shared_logger.error(f"BasePreprocessor: data transfer: File not found: {file_path} — {e}")
            # except PermissionError as e:
                # shared_logger.error(f"BasePreprocessor: data transfer: Permission denied: {file_path} — {e}")
            # except OSError as e:
                # shared_logger.error(f"BasePreprocessor: data transfer: OS error while accessing {file_path}: {e}")
            # except Exception as e:
                # shared_logger.exception(f"BasePreprocessor: data transfer: Unexpected error while reading {file_path}: {e}")
                # raise
                
    def matched_file_list(self, origin_path: Path, file_part : str) -> list[Path]:

        """
        Return a list of files from the source directory that match the timstamp of the :path: file.
        """
        # # Set of all files in the directory
        # files_in_dir = self.list_files_in_directory(origin_path.parent)
        # shared_logger.debug(f"BasePreprocessor: files_in_dir:  {files_in_dir}")
        
        # # Set the timestamp regular expression
        # filestamp = r"(\d{4}-\d{2}-\d{2}_\d{2}-\d{2}-\d{2}_\d+)"  # defined in the Resonate processing when they save the files.
        # # Match the filename timestamps to the input filename
        # matched = self.match_timestamp(file_part, files_in_dir, filestamp)
        # shared_logger.debug(f"BasePreprocessor: Matched files: {matched}")
        # # Add back the directory
        # matched_with_dir = [origin_path.parent / f for f in matched]
        # # Save the list of matched filenames to extra_files list to be used
        # # in the save() method. Conver t he strings to Path objects        
        # path_objects = [Path(p) for p in matched_with_dir]
        # return path_objects
        return []
        
            
    def save(
        self,
        path: Path | str,
        **kwargs: Any,
    ) -> None:
        """
        Save the data files from the IMU
        The bin file is the original data from the Phenomate IMU system
        The two CSV files are the raw positioning data and the GPS referenced data (GNSS data)
        """
        fpath = Path(path)
        fpath.mkdir(parents=True, exist_ok=True)

        # current_year = str(datetime.now(timezone.utc).year)
        # phenomate_version = get_version()
        # start_time = time.time()

        # self.copy_extra_files(fpath)

        # # End timer
        # end_time = time.time()
        # # Print elapsed time
        # shared_logger.info(f"Write time (IMU data): {end_time - start_time:.4f} seconds")



import logging
from typing import Callable

# from phenomate_core.get_logging import shared_logger
shared_logger = logging.getLogger("celery")


def get_version() -> str:
    """Returns the version of the package or an empty string (if available methods do not obtain the string)."""
    try:
        from importlib.metadata import version

        return version("phenomate-core")
    except ImportError as e:
        shared_logger.warning(f"Cannot get __version__ string: {e}")

    try:
        from pathlib import Path

        import tomlib  # in Python 3.11+

        pyproject_path = Path(__file__).parent / "pyproject.toml"

        with pyproject_path.open("rb") as f:
            data = tomlib.load(f)
        return data["project"]["version"]

    except ImportError as e:
        shared_logger.warning(f"Cannot get __version__ string: {e}")

    return ""


def get_task_logger(name: str) -> logging.Logger:
    try:
        from celery.utils.log import get_task_logger as _gtl  # local import
        get_celery_logger: Callable[[str], logging.Logger] = _gtl  # type: ignore[assignment]
        return get_celery_logger(name)
    except Exception:
        return logging.getLogger(name)


from __future__ import annotations

import logging
import time
from pathlib import Path
from typing import Any
import shutil

from phenomate_core.preprocessing.base import BasePreprocessor

# shared_logger = logging.getLogger("celery")
from phenomate_core.get_version import get_task_logger
shared_logger = get_task_logger(__name__)

class RS3Preprocessor(BasePreprocessor[Path]):
    """
    Code based
    
    """
    
    def __init__(self, path: str | Path, in_ext: str = "bin", **kwargs: Any):
        super().__init__(path, in_ext)   # pass parameters up to Base


    def extract(self, **kwargs: Any) -> None:
        """
        This method reads the '.origin' file extended version of the .bin file selected and
        reads the source directory (the directory where the .bin file was selected in the
        Phenomate GUI) and looks for files of the same name and timestamp to select and stores
        this as a list of filenames in the self.images list, to be processed in the save() method.
        """

        dir_part = self.path.parent  # this is another Path type
        file_part = self.path.name  # this is a str

        # Read the path that was written to the self.path+'.origin' file        
        origin_path = self.open_origin_file()  
        # Select the matching files from the path by the timestamp of the self.path file
        path_objects = self.matched_file_list(origin_path, file_part)
        
        for file_path in path_objects:
            shared_logger.info(f"RS3Preprocessor data transfer: file_path.suffix:  {file_path.suffix}")
            if file_path.suffix == ".25O":
                self.extra_files.append(file_path)
            elif file_path.suffix == ".25P":
                self.extra_files.append(file_path)
                
        shared_logger.info(f"RS3Preprocessor data transfer: number of related files:  {len(self.extra_files)}")
        # self.images = path_objects

    def copy_extra_files(self, fpath: Path) -> None:
        """
        Extra files that are associated with the RS3 (IMU Basestation) .25B data will be copied 
        to the destination directory.
        
        :fpath Path: is the directory in which to save the files
        
        This method impicitly uses the extra_files list that should be populated in the extract() method using
        open_origin_file() and matched_file_list() (see: RS3Preprocessor.extract())
        
        
        """
        for file_path in self.extra_files:
            try:
                # For the RS3 there should be 3 files .25B, .25O, .25P 
                if file_path.suffix == ".25O":
                    file_path_name_ext = fpath / self.get_output_name(
                        index=None, ext="25O", details=None
                    )
                elif file_path.suffix == ".25P":
                    file_path_name_ext = fpath / self.get_output_name(
                        index=None, ext="25P", details=None
                    )
                    
                    
                shutil.copy(file_path, file_path_name_ext)
                shared_logger.info(f"RS3Preprocessor.copy_extra_files(): RS3 data transfer: Copied file: {file_path_name_ext}")

            except FileNotFoundError as e:
                shared_logger.error(f"RS3Preprocessor: data transfer: File not found: {file_path} — {e}")
            except PermissionError as e:
                shared_logger.error(f"RS3Preprocessor: data transfer: Permission denied: {file_path} — {e}")
            except OSError as e:
                shared_logger.error(f"RS3Preprocessor: data transfer: OS error while accessing {file_path}: {e}")
            except Exception as e:
                shared_logger.exception(f"RS3Preprocessor: data transfer: Unexpected error while reading {file_path}: {e}")
                raise
                
    def matched_file_list(self, origin_path: Path, file_part : str) -> list[Path]:

        """
        Return a list of files from the source directory that match the timstamp of the :path: file.
        """
        # Set of all files in the directory
        files_in_dir = self.list_files_in_directory(origin_path.parent)
        shared_logger.info(f"RS3Preprocessor: files_in_dir:  {files_in_dir}")
        
        # Set the timestamp regular expression
        filestamp = r'(\d{14})'  # Added to the *end* of the file as RS3 data is downloaded seperately
        # Match the filename timestamps to the input filename
        matched = self.match_timestamp(file_part, files_in_dir, filestamp)
        shared_logger.info(f"RS3Preprocessor: Matched files: {matched}")
        # Add back the directory
        matched_with_dir = [origin_path.parent / f for f in matched]
        # Save the list of matched filenames to extra_files list to be used
        # in the save() method. Conver t he strings to Path objects        
        path_objects = [Path(p) for p in matched_with_dir]
        return path_objects
        
            
    def save(
        self,
        path: Path | str,
        **kwargs: Any,
    ) -> None:
        """
        Save the data files from the RS3 Base station
        The first file is ".25B", which s copied by the backend to the output directry.
        The following ".25O" and ".25P" files are copied using the "matched_files_list()" method
        """
        fpath = Path(path)
        fpath.mkdir(parents=True, exist_ok=True)

        # current_year = str(datetime.now(timezone.utc).year)
        # phenomate_version = get_version()
        start_time = time.time()

        self.copy_extra_files(fpath)

        # End timer
        end_time = time.time()
        # Print elapsed time
        shared_logger.info(f"RS3Preprocessor: Write time: {end_time - start_time:.4f} seconds")



import abc
import os
import re
import logging
from pathlib import Path
from typing import Any, Generic, TypeVar
import shutil
from datetime import datetime

import numpy as np
from numpy.typing import NDArray

T = TypeVar("T")

from ..get_version import get_task_logger
shared_logger = get_task_logger(__name__)

class BasePreprocessor(Generic[T], abc.ABC):
    def __init__(self, path: str | Path, in_ext: str = "bin", **kwargs: Any) -> None:
        self._in_ext = self.process_ext(in_ext)
        self._base_name = self.validate_file_path(path)
        self.images: list[T] = []
        self.extra_files: list[Path] = []
        self.system_timestamps: list[int] = []

    @staticmethod
    def bytes_to_numpy(image: bytes) -> NDArray[np.uint8]:
        return np.frombuffer(image, dtype=np.uint8)

    def validate_file_path(self, path: str | Path) -> str:
        fpath = Path(path)
        if not fpath.exists():
            raise FileNotFoundError(f"File doesn't exist: {fpath!s}")
        if not fpath.is_file():
            raise ValueError(f"Not a file: {fpath!s}")
        self._path = fpath
        name = self.path.name
        # Special processing for IMUU GNSS files
        if not name.endswith("_GNSS.csv"):
            if not name.endswith("." + self._in_ext):
                raise ValueError(f"Expects input file with ext: {self._in_ext}. Input: {name}")
        return name[: -len("." + self._in_ext)]

    def process_ext(self, ext: str) -> str:
        return ext[1:] if ext.startswith(".") else ext

    def get_output_name(self, index: int | None, ext: str, details: str | None = None) -> str:
        base = f"{self._base_name}"
        if index is not None:
            base += f"-{index:020}"
        if details is not None:
            base += f"_{details}"
        return f"{base}.{ext}"

    @abc.abstractmethod
    def extract(self, **kwargs: Any) -> None: ...

    @abc.abstractmethod
    def save(self, path: Path | str, **kwargs: Any) -> None: ...

    @property
    def path(self) -> Path:
        return self._path
        
    def open_origin_file(self) -> Path :
        """
        Opens the :.origin: file that is saved to the output directory and that contains
        the path to the source directory, from which the :self.path: file originated
        from.       
        """
        # Read the path that was written to the .origin file
        # N.B. self.path is where the .ibn file has already been copied to.
        origin_line = ""
        origin_file = str(self.path) + ".origin" 

        # read the first line of the .origin file which stores the path of the
        # origin of 
        with Path.open(Path(origin_file), encoding="utf-8") as f:
            origin_line = f.readline()
            origin_line = origin_line.strip()

        origin_path = Path(origin_line)
        shared_logger.debug(f"BasePreprocessor: Contents of .origin file:  {origin_path}")
        return origin_path


    @abc.abstractmethod
    def matched_file_list(self, origin_path: Path, file_part : str) -> list[Path]: 
        """
        Return a list of files from the source directory that match the timstamp of the :path: file.
        
        This method should be overridden to define the behavior
                of 'matched_file_list' in the derived class
        """
        ...
                
    @abc.abstractmethod
    def copy_extra_files(self, fpath: Path) -> None:
        """
        Extra files that are associated with the .bin proto buffer data can be copied to the destination directory.
        
        :fpath Path: is the directory in which to save the files
        
        This method impicitly uses the extra_files list that should be populated in the extract() method using
        open_origin_file() and matched_file_list() (see: ImuPreprocessor.extract())
        
        
        This method should be overridden to define the behavior
                of 'copy_extra_files' in the derived class

        """
        ...
       

    def return_closest_in_time(self, json_files : list, file_part_converted : str) -> str | None:
        """
        Returns the string from the list that is closest in time as the file_part_converted 
        string using the embedded timestamp data in the filenames.
        
        Returns: The string that has the closest timestamp
        
        Expects input timestamp format of: YYYY-MM-DD_HH-MM-SS_ddddddd
        
        """
        
        # The pattern string should be added to the app .envand .env.production environment files
        pattern = re.compile( r'(?P<date>\d{4}-\d{2}-\d{2})_(?P<time>\d{2}-\d{2}-\d{2})(?:_(?P<subsec>\d{6}))?' )
        
        # Convert json list to tuples (datetime, filename)
        json_times = [(self.extract_datetime(f, pattern), f) for f in json_files]
        
        has_none = any(x is None for x in json_times)
        if has_none:
            return None
        
        # Sort JSON files by time
        json_times.sort(key=lambda x: x[0])

        file_part_time = self.extract_datetime(file_part_converted, pattern)
        older_json = None
        for jt, jf in json_times:
            if jt < file_part_time:
                older_json = jf
            else:
                break
                
        return older_json
        
                    
    def extract_datetime(self, filename: str, pattern: re.Pattern) -> datetime | None:
        """
        Expects input timestamp format of: YYYY-MM-DD_HH-MM-SS_ddddddd
        
        """

        m = pattern.search(filename)
        if not m:
            raise ValueError(f"BasePreprocessor: extract_datetime: No date/time found")

        date_str = m.group("date")        # "2025-11-13"
        time_str = m.group("time")        # "14-05-33"
        subsec   = m.group("subsec")      # "1234567" or None

        date_str = date_str + '_' + time_str + '.' + subsec

        # date_str = filename.split("_")[0] + "_" + filename.split("_")[1] + "." + filename.split("_")[2]
        try:
            resdate = datetime.strptime(date_str, "%Y-%m-%d_%H-%M-%S.%f")
            return resdate
        except ValueError as e:
            shared_logger.error(f"BasePreprocessor: extract_datetime: {filename} — {e}")
            return None
             
        
    def extract_timestamp(self, filename: str, pattern : str):
        # Extract the timestamp pattern from the filename
        shared_logger.debug(f"BasePreprocessor: extract_timestamp(): matching expression = {pattern } ")
        shared_logger.debug(f"BasePreprocessor: extract_timestamp(): filename = {filename} ")
        match = re.search(pattern , filename)
        shared_logger.info(f"BasePreprocessor: extract_timestamp(): match = {match} ")
        if match: 
            return match.group(1) 
        else:
            return None

    def match_timestamp(
        self,
        target_filename: str,
        list_of_filenames: list[str],
        filestamp: str = r"(\d{4}-\d{2}-\d{2}_\d{2}-\d{2}-\d{2}_\d+)"
    ) -> list[str]:
        
        target_timestamp = self.extract_timestamp(target_filename, filestamp)
        if not target_timestamp:
            return []

        # Return all filenames that contain the same timestamp
        return [f for f in list_of_filenames if target_timestamp in f]

    @staticmethod
    def list_files_in_directory(directory: Path) -> list[str]:
        # List all files in the directory
        try:
            return [f for f in os.listdir(directory) if Path.is_file(directory / f)]
        except FileNotFoundError:
            print(f"Directory not found: {directory}")
            return []
        except PermissionError:
            print(f"Permission denied to access: {directory}")
            return []
